@@ -1,6 +1,6 @@
+'use strict';
 var common = require('../common.js');
 var path = require('path');
-var v8 = require('v8');
 
 var bench = common.createBenchmark(main, {
   path: [
@@ -23,12 +23,6 @@ function main(conf) {
   var p = path.posix;
   var input = '' + conf.path;
 
-  // Force optimization before starting the benchmark
-  p.extname(input);
-  v8.setFlagsFromString('--allow_natives_syntax');
-  eval('%OptimizeFunctionOnNextCall(p.extname)');
-  p.extname(input);
-
   bench.start();
   for (var i = 0; i < n; i++) {
     p.extname(input);
