@@ -1,7 +1,6 @@
 'use strict';
 var common = require('../common.js');
 var path = require('path');
-var v8 = require('v8');
 
 var bench = common.createBenchmark(main, {
   props: [
@@ -22,12 +21,6 @@ function main(conf) {
     name: props[4] || '',
   };
 
-  // Force optimization before starting the benchmark
-  p.format(obj);
-  v8.setFlagsFromString('--allow_natives_syntax');
-  eval('%OptimizeFunctionOnNextCall(p.format)');
-  p.format(obj);
-
   bench.start();
   for (var i = 0; i < n; i++) {
     p.format(obj);
