@@ -617,7 +617,16 @@ Module._initPaths = function() {
     homeDir = process.env.HOME;
   }
 
-  var paths = [path.resolve(process.execPath, '..', '..', 'lib', 'node')];
+  // $PREFIX/lib/node, where $PREFIX is the root of the Node.js installation.
+  var prefixDir;
+  // process.execPath is $PREFIX/bin/node except on Windows where it is
+  // $PREFIX\node.exe.
+  if (isWindows) {
+    prefixDir = path.resolve(process.execPath, '..');
+  } else {
+    prefixDir = path.resolve(process.execPath, '..', '..');
+  }
+  var paths = [path.resolve(prefixDir, 'lib', 'node')];
 
   if (homeDir) {
     paths.unshift(path.resolve(homeDir, '.node_libraries'));
@@ -652,9 +661,8 @@ Module._preloadModules = function(requests) {
       throw e;
     }
   }
-  requests.forEach(function(request) {
-    parent.require(request);
-  });
+  for (var n = 0; n < requests.length; n++)
+    parent.require(requests[n]);
 };
 
 Module._initPaths();
