@@ -1,7 +1,9 @@
 'use strict';
 
 const binding = process.binding('os');
+const getCPUs = binding.getCPUs;
 const getLoadAvg = binding.getLoadAvg;
+const pushValToArrayMax = process.binding('util').pushValToArrayMax;
 const constants = process.binding('constants').os;
 const internalUtil = require('internal/util');
 const isWindows = process.platform === 'win32';
@@ -10,7 +12,6 @@ exports.hostname = binding.getHostname;
 exports.uptime = binding.getUptime;
 exports.freemem = binding.getFreeMem;
 exports.totalmem = binding.getTotalMem;
-exports.cpus = binding.getCPUs;
 exports.type = binding.getOSType;
 exports.release = binding.getOSRelease;
 exports.networkInterfaces = binding.getInterfaceAddresses;
@@ -23,6 +24,26 @@ exports.loadavg = function loadavg() {
   return [avgValues[0], avgValues[1], avgValues[2]];
 };
 
+const cpuValues = new Float64Array(6 * pushValToArrayMax);
+function addCPUInfo() {
+  for (var i = 0, c = 0; i < arguments.length; ++i, c += 6) {
+    this[this.length] = {
+      model: arguments[i],
+      speed: cpuValues[c],
+      times: {
+        user: cpuValues[c + 1],
+        nice: cpuValues[c + 2],
+        sys: cpuValues[c + 3],
+        idle: cpuValues[c + 4],
+        irq: cpuValues[c + 5]
+      }
+    };
+  }
+}
+exports.cpus = function cpus() {
+  return getCPUs(addCPUInfo, cpuValues, []);
+};
+
 Object.defineProperty(exports, 'constants', {
   configurable: false,
   enumerable: true,
