@@ -7,10 +7,11 @@
 'use strict';
 
 const kHistorySize = 30;
+const kMincrlfDelay = 100;
+const kMaxcrlfDelay = 2000;
 
 const util = require('util');
 const debug = util.debuglog('readline');
-const internalUtil = require('internal/util');
 const inherits = util.inherits;
 const Buffer = require('buffer').Buffer;
 const EventEmitter = require('events');
@@ -22,40 +23,37 @@ const stripVTControlCharacters = internalReadline.stripVTControlCharacters;
 
 
 exports.createInterface = function(input, output, completer, terminal) {
-  var rl;
-  if (arguments.length === 1) {
-    rl = new Interface(input);
-  } else {
-    rl = new Interface(input, output, completer, terminal);
-  }
-  return rl;
+  return new Interface(input, output, completer, terminal);
 };
 
 
 function Interface(input, output, completer, terminal) {
   if (!(this instanceof Interface)) {
-    // call the constructor preserving original number of arguments
-    const self = Object.create(Interface.prototype);
-    Interface.apply(self, arguments);
-    return self;
+    return new Interface(input, output, completer, terminal);
   }
 
-  this._sawReturn = false;
+  this._sawReturnAt = 0;
   this.isCompletionEnabled = true;
+  this._sawKeyPress = false;
+  this._previousKey = null;
 
   EventEmitter.call(this);
   var historySize;
+  var removeHistoryDuplicates = false;
+  let crlfDelay;
   let prompt = '> ';
 
-  if (arguments.length === 1) {
+  if (input && input.input) {
     // an options object was given
     output = input.output;
     completer = input.completer;
     terminal = input.terminal;
     historySize = input.historySize;
+    removeHistoryDuplicates = input.removeHistoryDuplicates;
     if (input.prompt !== undefined) {
       prompt = input.prompt;
     }
+    crlfDelay = input.crlfDelay;
     input = input.input;
   }
 
@@ -84,6 +82,9 @@ function Interface(input, output, completer, terminal) {
   this.output = output;
   this.input = input;
   this.historySize = historySize;
+  this.removeHistoryDuplicates = !!removeHistoryDuplicates;
+  this.crlfDelay = Math.max(kMincrlfDelay,
+                            Math.min(kMaxcrlfDelay, crlfDelay >>> 0));
 
   // Check arity, 2 - for async, 1 for sync
   if (typeof completer === 'function') {
@@ -117,6 +118,14 @@ function Interface(input, output, completer, terminal) {
 
   function onkeypress(s, key) {
     self._ttyWrite(s, key);
+    if (key && key.sequence) {
+      // if the key.sequence is half of a surrogate pair
+      // (>= 0xd800 and <= 0xdfff), refresh the line so
+      // the character is displayed appropriately.
+      const ch = key.sequence.codePointAt(0);
+      if (ch >= 0xd800 && ch <= 0xdfff)
+        self._refreshLine();
+    }
   }
 
   function onresize() {
@@ -247,7 +256,16 @@ Interface.prototype._addHistory = function() {
   // if the history is disabled then return the line
   if (this.historySize === 0) return this.line;
 
+  // if the trimmed line is empty then return the line
+  if (this.line.trim().length === 0) return this.line;
+
   if (this.history.length === 0 || this.history[0] !== this.line) {
+    if (this.removeHistoryDuplicates) {
+      // Remove older history line if identical to new one
+      const dupIndex = this.history.indexOf(this.line);
+      if (dupIndex !== -1) this.history.splice(dupIndex, 1);
+    }
+
     this.history.unshift(this.line);
 
     // Only store so many
@@ -341,9 +359,10 @@ Interface.prototype._normalWrite = function(b) {
     return;
   }
   var string = this._decoder.write(b);
-  if (this._sawReturn) {
+  if (this._sawReturnAt &&
+      Date.now() - this._sawReturnAt <= this.crlfDelay) {
     string = string.replace(/^\n/, '');
-    this._sawReturn = false;
+    this._sawReturnAt = 0;
   }
 
   // Run test() on the new string chunk, not on the entire line buffer.
@@ -354,16 +373,15 @@ Interface.prototype._normalWrite = function(b) {
     this._line_buffer = null;
   }
   if (newPartContainsEnding) {
-    this._sawReturn = string.endsWith('\r');
+    this._sawReturnAt = string.endsWith('\r') ? Date.now() : 0;
 
     // got one or more newlines; process into "line" events
     var lines = string.split(lineEnding);
-    // either '' or (concievably) the unfinished portion of the next line
+    // either '' or (conceivably) the unfinished portion of the next line
     string = lines.pop();
     this._line_buffer = string;
-    lines.forEach(function(line) {
-      this._onLine(line);
-    }, this);
+    for (var n = 0; n < lines.length; n++)
+      this._onLine(lines[n]);
   } else if (string) {
     // no newlines this time, save what we have for next time
     this._line_buffer = string;
@@ -392,7 +410,7 @@ Interface.prototype._insertString = function(c) {
   }
 };
 
-Interface.prototype._tabComplete = function() {
+Interface.prototype._tabComplete = function(lastKeypressWasTab) {
   var self = this;
 
   self.pause();
@@ -408,9 +426,7 @@ Interface.prototype._tabComplete = function() {
     const completeOn = rv[1];  // the text that was completed
     if (completions && completions.length) {
       // Apply/show completions.
-      if (completions.length === 1) {
-        self._insertString(completions[0].slice(completeOn.length));
-      } else {
+      if (lastKeypressWasTab) {
         self._writeToOutput('\r\n');
         var width = completions.reduce(function(a, b) {
           return a.length > b.length ? a : b;
@@ -419,9 +435,9 @@ Interface.prototype._tabComplete = function() {
         if (!maxColumns || maxColumns === Infinity) {
           maxColumns = 1;
         }
-        var group = [], c;
+        var group = [];
         for (var i = 0, compLen = completions.length; i < compLen; i++) {
-          c = completions[i];
+          var c = completions[i];
           if (c === '') {
             handleGroup(self, group, width, maxColumns);
             group = [];
@@ -430,16 +446,15 @@ Interface.prototype._tabComplete = function() {
           }
         }
         handleGroup(self, group, width, maxColumns);
-
-        // If there is a common prefix to all matches, then apply that
-        // portion.
-        var f = completions.filter(function(e) { if (e) return e; });
-        var prefix = commonPrefix(f);
-        if (prefix.length > completeOn.length) {
-          self._insertString(prefix.slice(completeOn.length));
-        }
-
       }
+
+      // If there is a common prefix to all matches, then apply that portion.
+      const f = completions.filter(function(e) { if (e) return e; });
+      const prefix = commonPrefix(f);
+      if (prefix.length > completeOn.length) {
+        self._insertString(prefix.slice(completeOn.length));
+      }
+
       self._refreshLine();
     }
   });
@@ -447,7 +462,7 @@ Interface.prototype._tabComplete = function() {
 
 // this = Interface instance
 function handleGroup(self, group, width, maxColumns) {
-  if (group.length == 0) {
+  if (group.length === 0) {
     return;
   }
   var minRows = Math.ceil(group.length / maxColumns);
@@ -472,14 +487,15 @@ function handleGroup(self, group, width, maxColumns) {
 }
 
 function commonPrefix(strings) {
-  if (!strings || strings.length == 0) {
+  if (!strings || strings.length === 0) {
     return '';
   }
+  if (strings.length === 1) return strings[0];
   var sorted = strings.slice().sort();
   var min = sorted[0];
   var max = sorted[sorted.length - 1];
   for (var i = 0, len = min.length; i < len; i++) {
-    if (min[i] != max[i]) {
+    if (min[i] !== max[i]) {
       return min.slice(0, i);
     }
   }
@@ -689,10 +705,12 @@ Interface.prototype._moveCursor = function(dx) {
 
 // handle a write from the tty
 Interface.prototype._ttyWrite = function(s, key) {
+  const previousKey = this._previousKey;
   key = key || {};
+  this._previousKey = key;
 
   // Ignore escape key - Fixes #2876
-  if (key.name == 'escape') return;
+  if (key.name === 'escape') return;
 
   if (key.ctrl && key.shift) {
     /* Control and shift pressed */
@@ -773,7 +791,7 @@ Interface.prototype._ttyWrite = function(s, key) {
         break;
 
       case 'z':
-        if (process.platform == 'win32') break;
+        if (process.platform === 'win32') break;
         if (this.listenerCount('SIGTSTP') > 0) {
           this.emit('SIGTSTP');
         } else {
@@ -842,20 +860,22 @@ Interface.prototype._ttyWrite = function(s, key) {
     /* No modifier keys used */
 
     // \r bookkeeping is only relevant if a \n comes right after.
-    if (this._sawReturn && key.name !== 'enter')
-      this._sawReturn = false;
+    if (this._sawReturnAt && key.name !== 'enter')
+      this._sawReturnAt = 0;
 
     switch (key.name) {
       case 'return':  // carriage return, i.e. \r
-        this._sawReturn = true;
+        this._sawReturnAt = Date.now();
         this._line();
         break;
 
       case 'enter':
-        if (this._sawReturn)
-          this._sawReturn = false;
-        else
+        // When key interval > crlfDelay
+        if (this._sawReturnAt === 0 ||
+            Date.now() - this._sawReturnAt > this.crlfDelay) {
           this._line();
+        }
+        this._sawReturnAt = 0;
         break;
 
       case 'backspace':
@@ -893,7 +913,8 @@ Interface.prototype._ttyWrite = function(s, key) {
       case 'tab':
         // If tab completion enabled, do that...
         if (typeof this.completer === 'function' && this.isCompletionEnabled) {
-          this._tabComplete();
+          const lastKeypressWasTab = previousKey && previousKey.name === 'tab';
+          this._tabComplete(lastKeypressWasTab);
           break;
         }
         // falls through
@@ -946,6 +967,10 @@ function emitKeypressEvents(stream, iface) {
       if (r) {
         clearTimeout(timeoutId);
 
+        if (iface) {
+          iface._sawKeyPress = r.length === 1;
+        }
+
         for (var i = 0; i < r.length; i++) {
           if (r[i] === '\t' && typeof r[i + 1] === 'string' && iface) {
             iface.isCompletionEnabled = false;
@@ -978,7 +1003,7 @@ function emitKeypressEvents(stream, iface) {
   }
 
   function onNewListener(event) {
-    if (event == 'keypress') {
+    if (event === 'keypress') {
       stream.on('data', onData);
       stream.removeListener('newListener', onNewListener);
     }
@@ -1075,38 +1100,3 @@ function clearScreenDown(stream) {
   stream.write('\x1b[0J');
 }
 exports.clearScreenDown = clearScreenDown;
-
-
-/**
- * Returns the Unicode code point for the character at the
- * given index in the given string. Similar to String.charCodeAt(),
- * but this function handles surrogates (code point >= 0x10000).
- */
-
-function codePointAt(str, index) {
-  var code = str.charCodeAt(index);
-  var low;
-  if (0xd800 <= code && code <= 0xdbff) { // High surrogate
-    low = str.charCodeAt(index + 1);
-    if (!isNaN(low)) {
-      code = 0x10000 + (code - 0xd800) * 0x400 + (low - 0xdc00);
-    }
-  }
-  return code;
-}
-exports.codePointAt = internalUtil.deprecate(codePointAt,
-    'readline.codePointAt is deprecated. ' +
-    'Use String.prototype.codePointAt instead.');
-
-
-exports.getStringWidth = internalUtil.deprecate(getStringWidth,
-    'getStringWidth is deprecated and will be removed.');
-
-
-exports.isFullWidthCodePoint = internalUtil.deprecate(isFullWidthCodePoint,
-    'isFullWidthCodePoint is deprecated and will be removed.');
-
-
-exports.stripVTControlCharacters = internalUtil.deprecate(
-    stripVTControlCharacters,
-    'stripVTControlCharacters is deprecated and will be removed.');
