@@ -1,15 +1,7 @@
 'use strict';
 
-function importPunycode() {
-  try {
-    return process.binding('icu');
-  } catch (e) {
-    return require('punycode');
-  }
-}
-
-const { toASCII } = importPunycode();
-
+const { toASCII } = process.binding('config').hasIntl ?
+  process.binding('icu') : require('punycode');
 const { StorageObject, hexTable } = require('internal/querystring');
 const internalUrl = require('internal/url');
 exports.parse = urlParse;
@@ -84,7 +76,7 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
   return u;
 }
 
-Url.prototype.parse = function(url, parseQueryString, slashesDenoteHost) {
+Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
   if (typeof url !== 'string') {
     throw new TypeError('Parameter "url" must be a string, not ' + typeof url);
   }
@@ -319,7 +311,10 @@ Url.prototype.parse = function(url, parseQueryString, slashesDenoteHost) {
       // It only converts parts of the domain name that
       // have non-ASCII characters, i.e. it doesn't matter if
       // you call it with a domain that already is ASCII-only.
-      this.hostname = toASCII(this.hostname);
+
+      // Use lenient mode (`true`) to try to support even non-compliant
+      // URLs.
+      this.hostname = toASCII(this.hostname, true);
     }
 
     var p = this.port ? ':' + this.port : '';
@@ -562,7 +557,7 @@ function urlFormat(obj, options) {
   return obj.format();
 }
 
-Url.prototype.format = function() {
+Url.prototype.format = function format() {
   var auth = this.auth || '';
   if (auth) {
     auth = encodeAuth(auth);
@@ -647,7 +642,7 @@ function urlResolve(source, relative) {
   return urlParse(source, false, true).resolve(relative);
 }
 
-Url.prototype.resolve = function(relative) {
+Url.prototype.resolve = function resolve(relative) {
   return this.resolveObject(urlParse(relative, false, true)).format();
 };
 
@@ -656,7 +651,7 @@ function urlResolveObject(source, relative) {
   return urlParse(source, false, true).resolveObject(relative);
 }
 
-Url.prototype.resolveObject = function(relative) {
+Url.prototype.resolveObject = function resolveObject(relative) {
   if (typeof relative === 'string') {
     var rel = new Url();
     rel.parse(relative, false, true);
@@ -934,7 +929,7 @@ Url.prototype.resolveObject = function(relative) {
   return result;
 };
 
-Url.prototype.parseHost = function() {
+Url.prototype.parseHost = function parseHost() {
   var host = this.host;
   var port = portPattern.exec(host);
   if (port) {
