@@ -2,6 +2,7 @@
 #include "v8.h"
 #include "env.h"
 #include "env-inl.h"
+#include "string_bytes.h"
 
 #include <errno.h>
 #include <string.h>
@@ -16,7 +17,7 @@
 # include <unistd.h>        // gethostname, sysconf
 # include <sys/param.h>     // MAXHOSTNAMELEN on Linux and the BSDs.
 # include <sys/utsname.h>
-#endif  // __MINGW32__
+#endif  // __POSIX__
 
 // Add Windows fallback.
 #ifndef MAXHOSTNAMELEN
@@ -27,11 +28,15 @@ namespace node {
 namespace os {
 
 using v8::Array;
+using v8::ArrayBuffer;
 using v8::Boolean;
 using v8::Context;
+using v8::Float64Array;
+using v8::Function;
 using v8::FunctionCallbackInfo;
 using v8::Integer;
 using v8::Local;
+using v8::Null;
 using v8::Number;
 using v8::Object;
 using v8::String;
@@ -67,7 +72,7 @@ static void GetOSType(const FunctionCallbackInfo<Value>& args) {
   }
   rval = info.sysname;
 #else  // __MINGW32__
-  rval ="Windows_NT";
+  rval = "Windows_NT";
 #endif  // __POSIX__
 
   args.GetReturnValue().Set(OneByteString(env->isolate(), rval));
@@ -83,7 +88,14 @@ static void GetOSRelease(const FunctionCallbackInfo<Value>& args) {
   if (uname(&info) < 0) {
     return env->ThrowErrnoException(errno, "uname");
   }
+# ifdef _AIX
+  char release[256];
+  snprintf(release, sizeof(release),
+           "%s.%s", info.version, info.release);
+  rval = release;
+# else
   rval = info.release;
+# endif
 #else  // Windows
   char release[256];
   OSVERSIONINFOW info;
@@ -111,36 +123,47 @@ static void GetOSRelease(const FunctionCallbackInfo<Value>& args) {
 static void GetCPUInfo(const FunctionCallbackInfo<Value>& args) {
   Environment* env = Environment::GetCurrent(args);
   uv_cpu_info_t* cpu_infos;
-  int count, i;
+  int count, i, field_idx;
 
   int err = uv_cpu_info(&cpu_infos, &count);
   if (err)
     return;
 
-  Local<Array> cpus = Array::New(env->isolate());
-  for (i = 0; i < count; i++) {
+  CHECK(args[0]->IsFunction());
+  Local<Function> addfn = args[0].As<Function>();
+
+  CHECK(args[1]->IsFloat64Array());
+  Local<Float64Array> array = args[1].As<Float64Array>();
+  CHECK_EQ(array->Length(), 6 * NODE_PUSH_VAL_TO_ARRAY_MAX);
+  Local<ArrayBuffer> ab = array->Buffer();
+  double* fields = static_cast<double*>(ab->GetContents().Data());
+
+  CHECK(args[2]->IsArray());
+  Local<Array> cpus = args[2].As<Array>();
+
+  Local<Value> model_argv[NODE_PUSH_VAL_TO_ARRAY_MAX];
+  int model_idx = 0;
+
+  for (i = 0, field_idx = 0; i < count; i++) {
     uv_cpu_info_t* ci = cpu_infos + i;
 
-    Local<Object> times_info = Object::New(env->isolate());
-    times_info->Set(env->user_string(),
-                    Number::New(env->isolate(), ci->cpu_times.user));
-    times_info->Set(env->nice_string(),
-                    Number::New(env->isolate(), ci->cpu_times.nice));
-    times_info->Set(env->sys_string(),
-                    Number::New(env->isolate(), ci->cpu_times.sys));
-    times_info->Set(env->idle_string(),
-                    Number::New(env->isolate(), ci->cpu_times.idle));
-    times_info->Set(env->irq_string(),
-                    Number::New(env->isolate(), ci->cpu_times.irq));
+    fields[field_idx++] = ci->speed;
+    fields[field_idx++] = ci->cpu_times.user;
+    fields[field_idx++] = ci->cpu_times.nice;
+    fields[field_idx++] = ci->cpu_times.sys;
+    fields[field_idx++] = ci->cpu_times.idle;
+    fields[field_idx++] = ci->cpu_times.irq;
+    model_argv[model_idx++] = OneByteString(env->isolate(), ci->model);
 
-    Local<Object> cpu_info = Object::New(env->isolate());
-    cpu_info->Set(env->model_string(),
-                  OneByteString(env->isolate(), ci->model));
-    cpu_info->Set(env->speed_string(),
-                  Number::New(env->isolate(), ci->speed));
-    cpu_info->Set(env->times_string(), times_info);
+    if (model_idx >= NODE_PUSH_VAL_TO_ARRAY_MAX) {
+      addfn->Call(env->context(), cpus, model_idx, model_argv).ToLocalChecked();
+      model_idx = 0;
+      field_idx = 0;
+    }
+  }
 
-    (*cpus)->Set(i, cpu_info);
+  if (model_idx > 0) {
+    addfn->Call(env->context(), cpus, model_idx, model_argv).ToLocalChecked();
   }
 
   uv_free_cpu_info(cpu_infos, count);
@@ -173,14 +196,12 @@ static void GetUptime(const FunctionCallbackInfo<Value>& args) {
 
 
 static void GetLoadAvg(const FunctionCallbackInfo<Value>& args) {
-  Environment* env = Environment::GetCurrent(args);
-  double loadavg[3];
+  CHECK(args[0]->IsFloat64Array());
+  Local<Float64Array> array = args[0].As<Float64Array>();
+  CHECK_EQ(array->Length(), 3);
+  Local<ArrayBuffer> ab = array->Buffer();
+  double* loadavg = static_cast<double*>(ab->GetContents().Data());
   uv_loadavg(loadavg);
-  Local<Array> loads = Array::New(env->isolate(), 3);
-  loads->Set(0, Number::New(env->isolate(), loadavg[0]));
-  loads->Set(1, Number::New(env->isolate(), loadavg[1]));
-  loads->Set(2, Number::New(env->isolate(), loadavg[2]));
-  args.GetReturnValue().Set(loads);
 }
 
 
@@ -200,7 +221,7 @@ static void GetInterfaceAddresses(const FunctionCallbackInfo<Value>& args) {
   ret = Object::New(env->isolate());
 
   if (err == UV_ENOSYS) {
-    args.GetReturnValue().Set(ret);
+    return args.GetReturnValue().Set(ret);
   } else if (err) {
     return env->ThrowUVException(err, "uv_interface_addresses");
   }
@@ -217,7 +238,7 @@ static void GetInterfaceAddresses(const FunctionCallbackInfo<Value>& args) {
     name = OneByteString(env->isolate(), raw_name);
 #endif
 
-    if (ret->Has(name)) {
+    if (ret->Has(env->context(), name).FromJust()) {
       ifarr = Local<Array>::Cast(ret->Get(name));
     } else {
       ifarr = Array::New(env->isolate());
@@ -290,6 +311,72 @@ static void GetHomeDirectory(const FunctionCallbackInfo<Value>& args) {
 }
 
 
+static void GetUserInfo(const FunctionCallbackInfo<Value>& args) {
+  Environment* env = Environment::GetCurrent(args);
+  uv_passwd_t pwd;
+  enum encoding encoding;
+
+  if (args[0]->IsObject()) {
+    Local<Object> options = args[0].As<Object>();
+    Local<Value> encoding_opt = options->Get(env->encoding_string());
+    encoding = ParseEncoding(env->isolate(), encoding_opt, UTF8);
+  } else {
+    encoding = UTF8;
+  }
+
+  const int err = uv_os_get_passwd(&pwd);
+
+  if (err) {
+    return env->ThrowUVException(err, "uv_os_get_passwd");
+  }
+
+  Local<Value> uid = Number::New(env->isolate(), pwd.uid);
+  Local<Value> gid = Number::New(env->isolate(), pwd.gid);
+  Local<Value> username = StringBytes::Encode(env->isolate(),
+                                              pwd.username,
+                                              encoding);
+  Local<Value> homedir = StringBytes::Encode(env->isolate(),
+                                             pwd.homedir,
+                                             encoding);
+  Local<Value> shell;
+
+  if (pwd.shell == NULL)
+    shell = Null(env->isolate());
+  else
+    shell = StringBytes::Encode(env->isolate(), pwd.shell, encoding);
+
+  uv_os_free_passwd(&pwd);
+
+  if (username.IsEmpty()) {
+    return env->ThrowUVException(UV_EINVAL,
+                                 "uv_os_get_passwd",
+                                 "Invalid character encoding for username");
+  }
+
+  if (homedir.IsEmpty()) {
+    return env->ThrowUVException(UV_EINVAL,
+                                 "uv_os_get_passwd",
+                                 "Invalid character encoding for homedir");
+  }
+
+  if (shell.IsEmpty()) {
+    return env->ThrowUVException(UV_EINVAL,
+                                 "uv_os_get_passwd",
+                                 "Invalid character encoding for shell");
+  }
+
+  Local<Object> entry = Object::New(env->isolate());
+
+  entry->Set(env->uid_string(), uid);
+  entry->Set(env->gid_string(), gid);
+  entry->Set(env->username_string(), username);
+  entry->Set(env->homedir_string(), homedir);
+  entry->Set(env->shell_string(), shell);
+
+  args.GetReturnValue().Set(entry);
+}
+
+
 void Initialize(Local<Object> target,
                 Local<Value> unused,
                 Local<Context> context) {
@@ -304,6 +391,7 @@ void Initialize(Local<Object> target,
   env->SetMethod(target, "getOSRelease", GetOSRelease);
   env->SetMethod(target, "getInterfaceAddresses", GetInterfaceAddresses);
   env->SetMethod(target, "getHomeDirectory", GetHomeDirectory);
+  env->SetMethod(target, "getUserInfo", GetUserInfo);
   target->Set(FIXED_ONE_BYTE_STRING(env->isolate(), "isBigEndian"),
               Boolean::New(env->isolate(), IsBigEndian()));
 }
